/*
    Scanbot Barcode Scanner SDK Flutter Plugin

    Copyright (c) 2020 doo GmbH. All rights reserved.
    https://scanbot.io/sdk
 */
package io.scanbot.barcode.sdk.flutter

import android.app.Activity
import android.app.Application
import android.content.Context
import android.net.Uri
import android.util.Log
import androidx.annotation.NonNull
import com.google.gson.FieldNamingPolicy
import com.google.gson.GsonBuilder
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.scanbot.barcode.sdk.flutter.utils.StorageUtils
import io.scanbot.barcode.sdk.flutter.utils.getBitmapFromContentUri
import io.scanbot.barcode.sdk.flutter.utils.getBitmapFromFileUri
import io.scanbot.sap.Status.StatusOkay
import io.scanbot.sap.Status.StatusTrial
import io.scanbot.sdk.barcode.entity.*
import io.scanbot.sdk.barcode.entity.BarcodeFormat.valueOf
import io.scanbot.sdk.barcode_scanner.ScanbotBarcodeScannerSDK
import io.scanbot.sdk.barcode_scanner.ScanbotBarcodeScannerSDKInitializer
import io.scanbot.sdk.ui.barcode_scanner.view.barcode.BarcodeScannerActivity
import io.scanbot.sdk.ui.barcode_scanner.view.barcode.batch.BatchBarcodeScannerActivity
import kotlinx.coroutines.*
import java.io.File
import java.text.SimpleDateFormat
import java.util.*
import kotlin.coroutines.CoroutineContext
import kotlin.coroutines.resume
import kotlin.coroutines.resumeWithException
import kotlin.coroutines.suspendCoroutine


class ScanbotBarcodeSdkPlugin() : FlutterPlugin,
        ActivityAware,
        MethodCallHandler,
        CoroutineResultHandler(),
        CoroutineScope {
    override val coroutineContext: CoroutineContext
        get() = job + Dispatchers.IO

    companion object {
        @JvmStatic
        private val gson = GsonBuilder()
                .setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
                .registerTypeAdapter(kotlin.ByteArray::class.java, ByteArrayToBase64TypeAdapter())
                .create()

        @JvmStatic
        var batchBarcodeCallbackChannel: MethodChannel? = null

        @JvmStatic
        private suspend fun getFormattedData(barcodeItem: BarcodeItem) = suspendCoroutine<BarcodeFormattedData> {
            val json = gson.toJson(barcodeItem)
            batchBarcodeCallbackChannel?.invokeMethod("getFormattedData", json.toString(), object : Result {
                override fun success(result: Any?) {
                    val json = (result as BarcodeFormattedJson)
                    val data = BarcodeFormattedData((json["title"] as? String)
                            ?: barcodeItem.barcodeFormat.name,
                            (json["subtitle"] as? String) ?: barcodeItem.text ?: "");
                    it.resume(data)
                }

                override fun error(errorCode: String?, errorMessage: String?, errorDetails: Any?) {
                    it.resumeWithException(IllegalStateException(errorMessage))
                }

                override fun notImplemented() {
                    it.resumeWithException(IllegalStateException("Not yet implemented"))
                }
            })
        }
    }


    private var job: Job = Job()

    private var currentActivity: Activity? = null
    private var appContext: Context? = null

    override fun onDetachedFromActivity() {
        Log.d("ScanbotBarcodeSdkPlugin", "onDetachedFromActivity")
        currentActivity = null
    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
        val activity = binding.activity
        Log.d("ScanbotBarcodeSdkPlugin", "onReattachedToActivityForConfigChanges : ${activity.localClassName}")

        binding.addActivityResultListener(this)
        currentActivity = activity
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        val activity = binding.activity
        Log.d("ScanbotBarcodeSdkPlugin", "onAttachedToActivity : ${activity.localClassName}")
        binding.addActivityResultListener(this)
        currentActivity = activity
    }

    override fun onDetachedFromActivityForConfigChanges() {
        Log.d("ScanbotBarcodeSdkPlugin", "onDetachedFromActivityForConfigChanges")
        currentActivity = null
    }

    override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        appContext = flutterPluginBinding.applicationContext
        val channel = MethodChannel(flutterPluginBinding.binaryMessenger, "barcode_scanner")
        batchBarcodeCallbackChannel = MethodChannel(flutterPluginBinding.binaryMessenger, "batch_barcode_data_mapper")
        channel.setMethodCallHandler(this);
        Log.d("ScanbotBarcodeSdkPlugin", "onAttachedToEngine")
    }

    override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        job.cancel()
        Log.d("ScanbotBarcodeSdkPlugin", "onDetachedFromEngine")
    }


    override fun onMethodCall(call: MethodCall, result: Result) {
        Log.d("ScanbotBarcodeSdkPlugin", call.method)
        launch {
            when (call.method) {
                "initScanbotSdk" -> {
                    appContext?.let { initScanbotSdk(it as Application, call, result) }
                }
                "getLicenseStatus" -> {
                    currentActivity?.let { getLicenseStatus(it, call, result) }
                }
                "cleanupBarcodeStorage" -> {
                    currentActivity?.let { cleanupBarcodeStorage(it, call, result) }
                }
                "startBarcodeScanner" -> {
                    currentActivity?.let { proceedWithBarcodeScanner(it, call, result) }
                }
                "startBatchBarcodeScanner" -> {
                    currentActivity?.let { proceedWithBatchBarcodeScanner(it, call, result) }
                }
                "detectFromImageFile" -> {
                    currentActivity?.let { detectFromImageFile(it, call, result) }
                }
            }
        }
    }

    private suspend fun initScanbotSdk(
            context: Application,
            call: MethodCall,
            methodResult: Result
    ) {

        withContext(Dispatchers.Main) {
            try {
                val licenseKey = call.argument<String>("licenseKey") ?: ""
                val storageBaseDirectory = call.argument<String>("storageBaseDirectory")
                val loggingEnabled = call.argument<Boolean>("loggingEnabled") ?: false
                val scanbotSDKInitializer = ScanbotBarcodeScannerSDKInitializer()

                if (!storageBaseDirectory.isNullOrEmpty()) {
                    val uri = Uri.parse(storageBaseDirectory)
                    val customStorageBaseDirectory = File(uri.path)
                    customStorageBaseDirectory.mkdirs()
                    if (customStorageBaseDirectory.isDirectory.not()) {
                        methodResult.error(
                                "IOException",
                                "Specified storageBaseDirectory is not valid: $uri",
                                null
                        )
                    }
                    scanbotSDKInitializer.sdkFilesDirectory(
                            context,
                            customStorageBaseDirectory
                    )
                    StorageUtils.setCustomStorageBaseDirectory(customStorageBaseDirectory)
                    Log.d(
                            "SDK",
                            "Using custom storage base directory: " + customStorageBaseDirectory.absolutePath
                    )
                }


                scanbotSDKInitializer
                        .withLogging(loggingEnabled)

                if (licenseKey.isNotEmpty()) {
                    scanbotSDKInitializer.license(context, licenseKey)
                }

                val initialize = scanbotSDKInitializer.initialize(context)

                when (initialize.status) {
                    StatusOkay -> methodResult.success("Scanbot SDK initialized.")

                    StatusTrial -> methodResult.success(
                            "Scanbot SDK initialized. Trial mode activated. You can now test all features for 60 seconds."
                    )
                    else -> {
                        methodResult.error("SDKNotInitialisedException",
                                "Scanbot SDK not initialised. license status = ${initialize.status.name}",
                                null
                        )
                    }
                }
            } catch (e: Throwable) {
                Log.e("ScanbotBarcodeSdkPlugin", e.toString())
                methodResult.error(e::class.java.simpleName, e.message, null)
            }
        }
    }

    private suspend fun getLicenseStatus(
            activity: Activity,
            call: MethodCall,
            methodResult: Result
    ) {
        withContext(Dispatchers.Main) {
            try {
                val scanbotSDK = ScanbotBarcodeScannerSDK(activity.application)
                val json = JsonObject()
                json.addProperty("isLicenseValid", scanbotSDK.licenseInfo.isValid)
                json.addProperty("status", scanbotSDK.licenseInfo.status.name)
                val time = scanbotSDK.licenseInfo.expirationDate?.time
                time?.let { date ->
                    val sdf: SimpleDateFormat = SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS")
                    sdf.timeZone = TimeZone.getTimeZone("CET")
                    json.addProperty("expirationDate", sdf.format(date))
                }
                methodResult.success(json.toString())
            } catch (e: Exception) {
                Log.e("ScanbotBarcodeSdkPlugin", e.toString())
                methodResult.error(
                        "Sdk exception",
                        e.localizedMessage,
                        null
                )
            }
        }
    }

    private suspend fun cleanupBarcodeStorage(
            activity: Activity,
            call: MethodCall,
            methodResult: Result
    ) {
        withContext(Dispatchers.Main) {
            try {
                val scanbotSDK = ScanbotBarcodeScannerSDK(activity.application)
                scanbotSDK.barcodeFileStorage().cleanupBarcodeImagesDirectory();
                methodResult.success("");
            } catch (e: Exception) {
                Log.e("ScanbotBarcodeSdkPlugin", e.toString())
                methodResult.error(
                        "Sdk exception",
                        e.localizedMessage,
                        null
                )
            }
        }
    }

    private suspend fun proceedWithBarcodeScanner(
            activity: Activity,
            call: MethodCall,
            methodResult: Result
    ) {
        val barcodeCameraConfiguration = try {
            readBarcodeScannerConfiguration(call, activity)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) { methodResult.error(e::class.java.simpleName, e.message, null) }
            return
        }
        val intent =
                BarcodeScannerActivity.newIntent(activity, barcodeCameraConfiguration)
        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }

        activityResult?.let {
            if (it.resultCode == Activity.RESULT_OK) {
                val barcodeData =
                        it.data?.getParcelableExtra<BarcodeScanningResult>(BarcodeScannerActivity.SCANNED_BARCODE_EXTRA)
                val imagePath = it.data?.getStringExtra(BarcodeScannerActivity.SCANNED_BARCODE_IMAGE_PATH_EXTRA)
                val previewPath = it.data?.getStringExtra(BarcodeScannerActivity.SCANNED_BARCODE_PREVIEW_FRAME_PATH_EXTRA)
                barcodeData?.let {
                    val result = gson.toJsonTree(it).asJsonObject
                    (imagePath ?: previewPath)?.let {
                        val uri = Uri.parse(it).toString()
                        result.addProperty("barcodeImageURI", uri)
                    }
                    withContext(Dispatchers.Main) { methodResult.success(gson.toJson(result)) }
                }
            } else {
                dispatchCanceled(methodResult)
            }
        } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
    }

    class FlutterInnerBarcodeFormatter : FormattedBarcodeDataMapper {
        override fun decodeFormattedData(barcodeItem: BarcodeItem): BarcodeFormattedData {
            try {
                return runBlocking {
                    withContext(Dispatchers.Main) { getFormattedData(barcodeItem) }
                }
            } catch (e: Throwable) {
                return BarcodeFormattedData(barcodeItem.barcodeFormat.name, barcodeItem.text)
            }

        }
    }

    private suspend fun proceedWithBatchBarcodeScanner(
            activity: Activity,
            call: MethodCall,
            methodResult: Result
    ) {
        val barcodeCameraConfiguration = try {
            readBatchBarcodeScannerConfiguration(call, activity)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) { methodResult.error(e::class.java.simpleName, e.message, null) }
            return
        }

        val arguments = call.arguments as BarcodeScannerConfigurationJson
        val isNeedToSetCallback = (arguments["shouldUseBarcodeFormatter"] as? Boolean) ?: false
        val intent = if (isNeedToSetCallback) {
            BatchBarcodeScannerActivity.newIntent(activity, barcodeCameraConfiguration, FlutterInnerBarcodeFormatter::class.java)
        } else {
            BatchBarcodeScannerActivity.newIntent(activity, barcodeCameraConfiguration)
        }


        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }

        activityResult?.let {
            if (it.resultCode == Activity.RESULT_OK) {
                val barcodeData =
                        it.data?.getParcelableExtra<BarcodeScanningResult>(BarcodeScannerActivity.SCANNED_BARCODE_EXTRA)
                val imagePath = it.data?.getStringExtra(BarcodeScannerActivity.SCANNED_BARCODE_IMAGE_PATH_EXTRA)
                val previewPath = it.data?.getStringExtra(BarcodeScannerActivity.SCANNED_BARCODE_PREVIEW_FRAME_PATH_EXTRA)
                barcodeData?.let {
                    val result = gson.toJsonTree(it).asJsonObject
                    (imagePath ?: previewPath)?.let {
                        val uri = Uri.parse(it).toString()
                        result.addProperty("barcodeImageURI", uri)
                    }
                    withContext(Dispatchers.Main) { methodResult.success(gson.toJson(result)) }
                }
            } else {
                dispatchCanceled(methodResult)
            }
        } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
    }


    private suspend fun detectFromImageFile(
            activity: Activity,
            call: MethodCall,
            methodResult: Result
    ) {
        //too add high sensitivity mode to and remove rotation
        withContext(Dispatchers.Default) {
            try {
                val scanbotBarcodeScannerSDK = ScanbotBarcodeScannerSDK(activity.application)
                val barcodeDetector = scanbotBarcodeScannerSDK.barcodeDetector()
                val uriString = call.argument<String>("uri")
                if (call.hasArgument("barcodeFormats")) {
                    val arguments = call.arguments as BarcodeDetectionParams
                    val barcodeFormats = arguments["barcodeFormats"] as ArrayList<String>
                    val formats = barcodeFormats.map { valueOf(it) }
                    barcodeDetector.setBarcodeFormatsFilter(formats)
                } else {
                    barcodeDetector.setBarcodeFormatsFilter(BarcodeFormat.values().toList())
                }
                if (call.hasArgument("enableHighSensitivityMode")) {
                    barcodeDetector.setupHighSensitivityMode(call.argument<Boolean>("enableHighSensitivityMode")
                            ?: false)
                }
                if (uriString.isNullOrEmpty()) {
                    methodResult.error("IllegalArgumentException", "Need to setup file uri =", null)
                }
                val uri = Uri.parse(uriString)
                val file = File(uri.path)
                val bitmap = when (uri.scheme) {
                    "content" -> {
                        getBitmapFromContentUri(activity.contentResolver, uri)
                    }
                    "file" -> {
                        getBitmapFromFileUri(file)
                    }
                    else -> {
                        if (file.exists()) {
                            getBitmapFromFileUri(file)
                        } else {
                            withContext(Dispatchers.Main) { methodResult.error("IllegalArgumentException", "unknown file path format", null) }
                            return@withContext
                        }
                    }
                }
                bitmap?.let {
                    val barcodeData = barcodeDetector.detectFromBitmap(it, 0)
                    barcodeData?.let {
                        val result = JsonObject()
                        it.also {
                            val barcodeItems = JsonArray()
                            it.barcodeItems.map { item ->
                                barcodeItems.add(JsonObject().apply {
                                    addProperty("text", item.text)
                                    addProperty("barcodeFormat", item.barcodeFormat.name)
                                })
                            }
                            result.add("barcodeItems", barcodeItems)
                        }
                        withContext(Dispatchers.Main) { methodResult.success(gson.toJson(result)) }
                    }
                }
                        ?: withContext(Dispatchers.Main) { methodResult.error("IllegalArgumentException", "bitmap not found", null) }

            } catch (e: Throwable) {
                withContext(Dispatchers.Main) { methodResult.error("IllegalStateException", e.message, null) }
            }
        }

    }


    private suspend fun dispatchCanceled(methodResult: Result) {
        withContext(Dispatchers.Main) {
            methodResult.error(
                    Activity.RESULT_CANCELED.toString(),
                    "Canceled",
                    null
            )
        }
    }

}

