/*
    Scanbot Barcode Scanner SDK Flutter Plugin

    Copyright (c) 2020 doo GmbH. All rights reserved.
    https://scanbot.io/sdk
 */
package io.scanbot.barcode.sdk.flutter

import android.app.Activity
import io.flutter.plugin.common.MethodCall
import io.scanbot.barcode.sdk.flutter.utils.ObjectMapper
import io.scanbot.sdk.barcode.entity.BarcodeFormat
import io.scanbot.sdk.ui.camera.FinderAspectRatio
import io.scanbot.sdk.ui.view.barcode.batch.configuration.BatchBarcodeScannerConfiguration
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeImageGenerationType
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeScannerConfiguration
import io.scanbot.sdk.ui.view.base.configuration.CameraOrientationMode


fun readBarcodeScannerConfiguration(
        call: MethodCall,
        activity: Activity
): BarcodeScannerConfiguration {
    val barcodeCameraConfiguration = BarcodeScannerConfiguration()
    val arguments = call.arguments as BarcodeScannerConfigurationJson
    val value = checkCameraOrientationParam(call, arguments)
    barcodeCameraConfiguration.setBarcodeImageGenerationType(BarcodeImageGenerationType.NONE)
    value?.let {
        barcodeCameraConfiguration.setOrientationLockMode(value)
    }
    val aspectRatio = checkFinderAspectRatioParam(call, arguments)
    aspectRatio?.let {
        barcodeCameraConfiguration.setFinderAspectRatio(it)
    }
    ObjectMapper.map(arguments, barcodeCameraConfiguration)

    if (arguments.containsKey("barcodeFormats")) {
        val barcodeFormats = arguments["barcodeFormats"] as ArrayList<String>
        val nativeBarcodeFormats = ArrayList<BarcodeFormat>()
        for (formatValue in barcodeFormats) {
            nativeBarcodeFormats.add(BarcodeFormat.valueOf(formatValue.replaceFirst("BarcodeFormat.", "")))
        }
        barcodeCameraConfiguration.setBarcodeFormatsFilter(nativeBarcodeFormats)
    }
    if (arguments.containsKey("barcodeImageGenerationType")) {
        val imageGenType = BarcodeImageGenerationType.valueOf(arguments["barcodeImageGenerationType"] as String)
        barcodeCameraConfiguration.setBarcodeImageGenerationType(imageGenType)
    }
    return barcodeCameraConfiguration
}

fun readBatchBarcodeScannerConfiguration(
        call: MethodCall,
        activity: Activity
): BatchBarcodeScannerConfiguration {
    val barcodeCameraConfiguration = BatchBarcodeScannerConfiguration()
    val arguments = call.arguments as BarcodeScannerConfigurationJson
    val orientationMode = checkCameraOrientationParam(call, arguments)
    barcodeCameraConfiguration.setBarcodeImageGenerationType(BarcodeImageGenerationType.NONE)
    orientationMode?.let {
        barcodeCameraConfiguration.setOrientationLockMode(orientationMode)
    }
    val aspectRatio = checkFinderAspectRatioParam(call, arguments)
    aspectRatio?.let {
        barcodeCameraConfiguration.setFinderAspectRatio(it)
    }
    ObjectMapper.map(arguments, barcodeCameraConfiguration)

    if (arguments.containsKey("barcodeFormats")) {
        val barcodeFormats = arguments["barcodeFormats"] as ArrayList<String>
        val nativeBarcodeFormats = ArrayList<BarcodeFormat>()
        for (formatValue in barcodeFormats) {
            nativeBarcodeFormats.add(BarcodeFormat.valueOf(formatValue.replaceFirst("BarcodeFormat.", "")))
        }
        barcodeCameraConfiguration.setBarcodeFormatsFilter(nativeBarcodeFormats)
    }
    if (arguments.containsKey("barcodeImageGenerationType")) {
        val imageGenType = BarcodeImageGenerationType.valueOf(arguments["barcodeImageGenerationType"] as String)
        barcodeCameraConfiguration.setBarcodeImageGenerationType(imageGenType)
    }
    return barcodeCameraConfiguration
}

private fun checkFinderAspectRatioParam(call: MethodCall, arguments: BarcodeScannerConfigurationJson): FinderAspectRatio? {
    val propertyName = "finderAspectRatio"
    if (arguments[propertyName] != null) {
        val aspRatio = arguments[propertyName] as? AspectRatioJson
        val width = aspRatio?.get("width") as? Int
        val height = aspRatio?.get("height") as? Int
        arguments.remove(propertyName)
        return FinderAspectRatio(width?.toDouble() ?: 1.0, height?.toDouble() ?: 1.0)
    }
    return null
}


private fun checkCameraOrientationParam(call: MethodCall, arguments: BarcodeScannerConfigurationJson): CameraOrientationMode? {
    if (call.hasArgument("orientationLockMode")) {

        val value = when (call.argument<String>("orientationLockMode")) {
            "PORTRAIT" -> CameraOrientationMode.PORTRAIT
            "PORTRAIT_UPSIDE_DOWN" -> CameraOrientationMode.PORTRAIT
            "LANDSCAPE_LEFT" -> CameraOrientationMode.LANDSCAPE
            "LANDSCAPE_RIGHT" -> CameraOrientationMode.LANDSCAPE
            "LANDSCAPE" -> CameraOrientationMode.LANDSCAPE
            else -> CameraOrientationMode.PORTRAIT
        }
        arguments.remove("orientationLockMode")
        return value
    }
    return null
}

